from qiskit.transpiler.preset_passmanagers import generate_preset_pass_manager
from qiskit import QuantumCircuit, QuantumRegister, ClassicalRegister
from typing import Literal, cast, List, Dict, Iterator, Tuple, Any
from qiskit_optimization.converters import QuadraticProgramToQubo
from qiskit_optimization.algorithms import MinimumEigenOptimizer
from qiskit_ibm_runtime import QiskitRuntimeService, Session
from qiskit_aer.primitives import Sampler as AerSampler
from qiskit.circuit import ParameterVector, Parameter
from qiskit_algorithms.utils import algorithm_globals
from qiskit_algorithms.optimizers import SPSA, COBYLA
from qiskit_ibm_runtime import SamplerV2 as Sampler
from qiskit_optimization import QuadraticProgram
from qiskit.compiler import transpile
from qiskit.providers import Backend
from qiskit.pulse import num_qubits
from qiskit_aer import AerSimulator
from qiskit_algorithms import QAOA
from sklearn.cluster import KMeans
from info_tools import InfoTools
from numpy import floating

import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
import random
import math
import sys


class QuantumSimulator:

    def __init__(self, service: Literal["aer", "ibm"] = "aer"):
        """
        :param service. Literal["aer"]. El servicio tecnológico con el cual se ejecuta la logica.
        """

        self.service: Literal["aer", "ibm"] = service
        self.sampler = None

        if self.service == "ibm":
            sys.exit(f"Se ha seleccionado el servicio {self.service} para ejecutar el simulador cuántico (usar: aer)")

        match self.service:

            case "aer":
                self.sampler = Sampler(AerSimulator())

    def run(self, qcs: List[QuantumCircuit], shots: int):
        """
        Metodo para ejecutar un simulador cuantico y obtener sus resultados
        :param qcs: Circuito cuantico que se quiere medir
        :param shots: Cantidad de veces que se ejecutara el circuito cuantico
        :return: Mediciones del circuito cuantico
        """

        # -- Definimos lista de resultados (homologado con metodo run de QuantumMachine)
        results: list = []

        # -- Definimos el sampler para ejecutar shots cantidad de veces el circuito cuantico especificado
        for qc in qcs:

            job = self.sampler.run([qc], shots=shots)

            # -- Lanzamos el job (tarea de ejecución del circuito cuántico) y obtenemos sus resultados
            job_result = job.result()[0].data.c
            results.append(job_result)

        # -- Obtenemos los resultados
        results = self.get_results(results)

        return results

    @staticmethod
    def get_results(results: List):
        """
        Obtenemos los resultados
        :param results: [List] Resultados
        :return: Lista de resultados
        """

        results_list: list = []

        # -- Accedemos a los valores de las mediciones del circuito cuantico
        for result in results:

            # -- Contamos la probabilidad de los resultados
            qc_ibm_results = result.get_counts()

            # -- Agregamos los resultados a lista de resultados
            results_list.append(qc_ibm_results)

        return results_list


class QuantumMachine:

    def __init__(self,
                 service: Literal["aer", "ibm"],
                 qm_api_key: str | None,
                 qm_connection_service: Literal["ibm_quantum", "ibm_cloud"] | None,
                 quantum_machine: Literal["ibm_brisbane", "ibm_kyiv", "ibm_sherbrooke", "least_busy"],
                 optimization_level: int = 1,
                 verbose: bool = True):
        """
        :param service. ["aer", "ibm"] El servicio tecnológico con el cual se ejecuta la lógica.
        :param qm_api_key. str | None. API KEY para conectarse con el servicio de computación cuántica de una empresa.
        :param qm_connection_service. Literal["ibm_quantum", "ibm_cloud"] | None. Servicio específico de computación cuántica. Por ejemplo, en el caso de IBM pueden ser a la fecha ibm_quantum | ibm_cloud
        :param quantum_machine. Literal["ibm_brisbane", "ibm_kyiv", "ibm_sherbrooke", "least_busy"]. Nombre del ordenador cuántico a utilizar. Por ejemplo, en el caso de IBM puede ser ibm_brisbane, ibm_kyiv, ibm_sherbrooke. Si se deja en least_busy,
        se buscará el ordenador menos ocupado para llevar a cabo la ejecución del algoritmo cuántico.
        :param optimization_level. Nivel de optimización del circuito cuántico
        """

        self.service: Literal["aer", "ibm"] = service
        self.qm_api_key: str | None = qm_api_key
        self.qm_connection_service: Literal["ibm_quantum", "ibm_cloud"] | None = qm_connection_service
        self.quantum_machine: Literal["ibm_brisbane", "ibm_kyiv", "ibm_sherbrooke", "least_busy"] = quantum_machine
        self.optimization_level: int = optimization_level
        self.verbose: bool = verbose

        # -- Instanciamos la clase para la consola
        self.IT = InfoTools()

        # -- Definimos la variable con la máquina elegida (puede ser igual a quantum_machine pero también la que resulte de least_busy)
        self.selected_machine: str | None = None
        self.ibm_machine_transpiler = None

        # -- Definimos el sampler de la máquina cuántica seleccionada
        self.sampler = None

        if self.service == "aer":
            self.IT.info_print(f"Se ha seleccionado el servicio {self.service} para ejecutar el ordenador cuantico (usar: ibm)", "light_red")
            sys.exit()

        match self.service:

            case "ibm":

                # -- Obtenemos el nombre de la máquina elegida y el transpilador de esa máquina
                if self.qm_connection_service == "ibm_quantum":

                    self.IT.sub_intro_print(f"Conectando con el servicio de computacion cuantica de IBM: ibm_quantum")

                    # -- Generamos el servicio de conexion
                    channel_type = Literal["ibm_cloud", "ibm_quantum", "local"]
                    channel = cast(channel_type, self.qm_connection_service)
                    self.service: QiskitRuntimeService = QiskitRuntimeService(channel=channel, token=self.qm_api_key)

                    # -- Chequeamos si la conexion ha sido exitosa
                    if self.service.active_account()["verify"]:
                        self.IT.info_print("Conexion realizada con exito")

                        self.IT.sub_intro_print("Datos de la cuenta")
                        _user_data: dict = self.service.usage()
                        _user_period_start: str = _user_data["period"]["start"]
                        _user_period_end: str = _user_data["period"]["end"]
                        _by_instance: str = _user_data["byInstance"][0]["instance"]
                        _user_quota: int = _user_data["byInstance"][0]["quota"]
                        _user_usage: int = _user_data["byInstance"][0]["usage"]
                        _user_pending_jobs: int = _user_data["byInstance"][0]["pendingJobs"]
                        _user_max_pending_jobs: int = _user_data["byInstance"][0]["maxPendingJobs"]

                        self.IT.info_print(f"Instancia de ejecucion: {_by_instance}")
                        self.IT.info_print(f"Cuota de ejecucion: {_user_quota}")
                        self.IT.info_print(f"Usos del usuario: {_user_usage}")
                        self.IT.info_print(f"Trabajos pendientes: {_user_pending_jobs} / {_user_max_pending_jobs}")

                        # -- Buscamos la maquina elegida
                        self.IT.sub_intro_print("Buscando ordenador cuantico...")
                        if self.quantum_machine == "least_busy":
                            self.IT.info_print(f"---> Buscando la maquina menos cargada...")
                            least_busy_machine = self.service.least_busy()
                            self.selected_machine = self.service.backend(least_busy_machine.name)
                            self.IT.info_print(f"---> La maquina menos cargada es: {self.selected_machine.name}")
                        else:
                            self.IT.info_print(f"---> Buscando la maquina {self.quantum_machine}...")
                            self.selected_machine: Backend = self.service.backend(self.quantum_machine)

                        self.IT.info_print(f"Numero de Qubits: {self.selected_machine.num_qubits}")
                        self.IT.info_print(f"Trabajos pendientes: {self.selected_machine.status().pending_jobs}")
                        self.IT.info_print(f"Operaciones permitidas: {self.selected_machine.operation_names}")
                        self.IT.info_print(f"Numero maximo de circuitos: {self.selected_machine.max_circuits}")

                        if self.selected_machine is not None:

                            # -- Generamos el transpilador o pass manager de la maquina elegida
                            self.connection_transpiler = generate_preset_pass_manager(backend=self.selected_machine, optimization_level=self.optimization_level)

                        else:
                            self.IT.info_print("El ordenador especificado no existe. FIN", "light_red")
                            sys.exit()

                elif self.qm_connection_service == "ibm_cloud":
                    pass

    def run(self, qc_list: List[QuantumCircuit], shots: int):

        """
        Metodo para ejecutar un ordenador cuantico y obtener sus resultados
        :param qc_list: List[QuantumCircuit]. Circuito cuantico que se quiere medir
        :param shots: int. Cantidad de veces que se ejecutara el circuito cuantico
        :return: Mediciones del circuito cuantico
        """

        # -- Transpilamos los circuitos antes de la ejecución
        transpiled_circuits = transpile(qc_list, backend=self.selected_machine,
                                        optimization_level=self.optimization_level)

        # -- Usamos el backend seleccionado para crear una sesión
        with Session(backend=self.selected_machine) as session:

            # -- Generamos el sampler de la máquina
            self.sampler = Sampler(mode=session)

            # -- Ejecutamos el trabajo con los circuitos transpilados y el número de shots
            job = self.sampler.run(transpiled_circuits, shots=shots)

            # -- Esperamos que el trabajo termine y obtenemos los resultados
            results = job.result()

        # -- Cerramos la sesión
        session.close()

        # -- Retornamos las cuasi-distribuciones de las mediciones
        if results is not None:
           results = self.get_results(results)
           return results

        else:
            self.IT.info_print("No se han podido obtener los resultados del ordenador cuántico. FIN", "light_red")
            sys.exit()

    @staticmethod
    def get_results(results: list):
        """
        Metodo para obtener la lista de resultados
        :param results: [list] Resultados a obtener
        :return: Lista de resultados
        """

        # -- Instaciamos una lista vacía para obtener los resultados
        results_list = []

        # -- Por cada resultado obtenemos la cantidad de repeticiones
        for result in results:
            counts = result.data.c.get_counts()
            results_list.append(counts)

        return results_list


class QuantumTechnology:

    def __init__(self,
                 quantum_technology: Literal["simulator", "quantum_machine"] = "simulator",
                 service: Literal["aer", "ibm"] = "aer",
                 qm_api_key: str | None = None,
                 qm_connection_service: Literal["ibm_quantum", "ibm_cloud"] | None = None,
                 quantum_machine: Literal["ibm_brisbane", "ibm_kyiv", "ibm_sherbrooke", "least_busy"] = "least_busy",
                 max_qubits: int = 32
                 ):

        """
        Metodo que instancia un objeto QuantumTechnology, el cual puede ser un simulador o un conector a una máquina cuántica real
        :param quantum_technology. Literal["simulator", "quantum_machine"]. Tecnología cuántica con la que calculan la lógica. Si es simulator, se hará con un simulador definido en el
        parámetro service. Si es quantum_machine, el algoritmo se ejecutará en una máquina cuántica definida en el parámetro technology.
        :param service. ["aer", "ibm"] El servicio tecnológico con el cual se ejecuta la lógica.
        :param qm_api_key. API KEY para conectarse con el servicio de computación cuántica de una empresa.
        :param qm_connection_service. Literal["ibm_quantum", "ibm_cloud"] | None. Servicio específico de computación cuántica. Por ejemplo, en el caso de IBM pueden ser a la fecha ibm_quantum | ibm_cloud
        :param quantum_machine. Literal["ibm_brisbane", "ibm_kyiv", "ibm_sherbrooke", "least_busy"]. Nombre del ordenador cuántico a utilizar. Por ejemplo, en el caso de IBM puede ser ibm_brisbane, ibm_kyiv, ibm_sherbrooke. Si se deja en least_busy,
        se buscará el ordenador menos ocupado para llevar a cabo la ejecución del algoritmo cuántico.
        """

        self.quantum_technology: Literal["simulator", "quantum_machine"] = quantum_technology
        self.service: Literal["aer", "ibm"] = service
        self.qm_api_key: str | None = qm_api_key
        self.qm_connection_service: Literal["ibm_quantum", "ibm_cloud"] | None = qm_connection_service
        self.quantum_machine: Literal["ibm_brisbane", "ibm_kyiv", "ibm_sherbrooke", "least_busy"] = quantum_machine
        self.max_qubits: int = max_qubits

        # -- TODO: Diccionario de tecnologias y servicios habilitados (actualizar periódicamente)
        self._allowed_quantum_tech: dict = {"simulator": ["aer"],
                                            "quantum_machine": ["ibm"],
                                            "quantum_machines": {"ibm": ["ibm_brisbane", "ibm_kyiv", "ibm_sherbrooke", "least_busy"]},
                                            "quantum_services": {"ibm": ["ibm_quantum", "ibm_cloud"]}}

        # -- Validamos que los parámetros relacionados con la tecnología cuántica son correctos
        self.validate_input_parameters()

        # -- Generamos el objeto que ejecuta el algoritmo cuántico
        self.execution_object: QuantumSimulator | QuantumMachine | None = None

        match quantum_technology:

            case "simulator":
                self.execution_object: QuantumSimulator = QuantumSimulator(self.service)

            case "quantum_machine":
                self.execution_object: QuantumMachine = QuantumMachine(self.service, self.qm_api_key, self.qm_connection_service, self.quantum_machine, 1)

    # ---------------------------- Metodos para tecnología cuántica  ------------------------------------- #

    def get_quantum_technology(self):
        """
        Metodo getter para retorna el execution object (objeto que ejecuta un circuito en un ordenador cuántico o en un simulador cuántico)
        :return: self.execution_object
        """
        return self.execution_object

    def validate_input_parameters(self) -> bool:
        """
        Metodo para validar los inputs que se han cargado en el constructor
        :return: True si todas las validaciones son correctas Excepction else
        """

        # -- Validamos la tecnologia cuántica definida
        if self.quantum_technology == "simulator":
            if self.service not in self._allowed_quantum_tech["simulator"]:
                raise ValueError(f"self.quantum_technology_executor: La randomness_technology escogida es {self.quantum_technology}. "
                                 f"Por tanto, debe estar entre los siguientes: {self._allowed_quantum_tech['simulator']}")

        # -- Validamos el ordenador cuántico elegido
        if self.quantum_technology == "quantum_machine":

            if self.service not in self._allowed_quantum_tech["quantum_machine"]:
                raise ValueError(f"self.technology: La technology escogida es {self.service}. "
                                 f"Por tanto, debe estar entre los siguientes: {self._allowed_quantum_tech['quantum_machine']}")

            if self.quantum_machine not in self._allowed_quantum_tech["quantum_machines"][f"{self.service}"]:
                raise ValueError(f"self.quantum_machine: La quantum_machine escogida es {self.quantum_machine}. "
                                 f"Por tanto, debe estar entre los siguientes: {self._allowed_quantum_tech['quantum_machines'][f'{self.service}']}")

            if self.qm_connection_service not in self._allowed_quantum_tech["quantum_services"][f"{self.service}"]:
                raise ValueError(f"self.qm_connection_service: El qm_connection_service escogido es {self.qm_connection_service}. "
                                 f"Por tanto, debe estar entre los siguientes: {self._allowed_quantum_tech['quantum_services'][f'{self.service}']}")

        return True

    @staticmethod
    def create_generator(bounds_dict: dict) -> Tuple[QuantumCircuit, ParameterVector]:
        """Metodo que crea un circuito cuántico parametrizado que actúa como generador"""

        # -- Definimos los parámetros a entrenar
        parameters = ParameterVector("θ", len(bounds_dict.keys()))

        # -- Definimos el circuito cuántico
        qc: QuantumCircuit = QuantumCircuit(len(bounds_dict.keys()), len(bounds_dict.keys()))

        # -- Generamos una estructura de reescalado a nivel horizontal
        for deep_level in range(1, 5):

            # -- Aplicamos rotación a la puertas parametrizada de cada qubit
            for i in range(len(bounds_dict.keys())):

                # -- Definimos el indice del qubit i para la rotación "Y" y "Z"
                ry_index = i
                rz_index = i

                # -- Aplicamos rotación en "Y"
                # qc.ry(parameters[ry_index] / deep_level, i)
                if i % 2 == 0:
                    qc.ry(parameters[ry_index] / deep_level, i)
                else:
                    qc.rz(parameters[rz_index] / deep_level, i)

            # -- Creamos entrelazamiento entre qubits adyacentes (CNOT between qubit i and qubit i+1)
            for i in range(len(bounds_dict.keys()) - 1):
                qc.cx(i, i + 1)

            # -- En algunos casos cerramos el circulo de entrelazamiento (CNOT entre el primero y el último qubit)
            if len(bounds_dict.keys()) > 2:
                qc.cx(control_qubit=0, target_qubit=len(bounds_dict.keys()) - 1)

        # -- Añadimos medidas a los qubits
        qc.measure(range(len(bounds_dict.keys())), range(len(bounds_dict.keys())))

        # -- Añadimos una barrera
        qc.barrier()

        return qc, parameters

    @staticmethod
    def create_generator_reproduction(bounds_dict: dict):
        # -- TODO: se podrían agregar qubits hacia abajo con alguna proporción de la función objetivo
        """Metodo que crea un circuito cuántico parametrizado que actúa como generador de la GAN"""

        # -- Definimos los parámetros a entrenar
        parameters = ParameterVector(name="θ", length=2 * len(bounds_dict.keys()))

        # -- Definimos el circuito cuántico
        qc: QuantumCircuit = QuantumCircuit(len(bounds_dict.keys()) + 1, len(bounds_dict.keys()) + 1)

        for deep_level in range(1, 5):

            # -- Aplicamos rotación a la puertas parametrizada de cada qubit
            for i in range(len(bounds_dict.keys())):

                # -- Aplicamos rotación en Y
                qc.ry(parameters[i] / deep_level, i)

                # -- Aplicamos rotación en Z
                qc.rz(parameters[i + len(bounds_dict.keys())] / deep_level, i)

            # -- Añadimos entrelazamiento entre los qubits con una puerta CNOT

            # -- Creamos entrelazamiento entre qubits adyacentes (CNOT between qubit i and qubit i+1)
            for i in range(len(bounds_dict.keys()) - 1):
                qc.cx(i, i + 1)

            # -- En algunos casos cerramos el circulo de entrelazamiento (CNOT between the first and last qubit)
            if len(bounds_dict.keys()) > 2:
                qc.cx(0, len(bounds_dict.keys()) - 1)

        parameter_metric = ParameterVector(name="metric", length=2)

        for deep_level in range(1, 5):

            # -- Aplicamos rotación a la puertas parametrizada de cada qubit
            for i in range(len(bounds_dict.keys()) - 1):

                # -- Aplicamos rotación en Y
                qc.ry(parameter_metric[i] / deep_level, len(bounds_dict.keys()))

                qc.rz(parameter_metric[i+1] / deep_level, len(bounds_dict.keys()))

            # -- Añadimos entrelazamiento entre los qubits con una puerta CNOT

            # -- Creamos entrelazamiento entre qubits adyacentes (CNOT between qubit i and qubit i+1)
            for i in range(0, len(bounds_dict.keys())):
                qc.cx(i, len(bounds_dict.keys()))

            # -- En algunos casos cerramos el circulo de entrelazamiento (CNOT between the first and last qubit)
            if len(bounds_dict.keys()) > 2:
                qc.cx(0, len(bounds_dict.keys()) - 1)

        # -- Añadimos medidas a los qubits
        qc.measure(range(len(bounds_dict.keys()) + 1), range(len(bounds_dict.keys()) + 1))

        # -- Añadimos una barrera
        qc.barrier()

        return qc, parameters, parameter_metric

    def generate_parameters(self, num_individuals: int, parameters: ParameterVector) -> List[List[float]]:

        """
        Metodo que genera las propiedades cuánticas de los individuos mediante circuitos cuánticos.
        Proceso: calcula cuántos qubits se necesitan por parámetro y ajusta valores según limitaciones malformantes.
        :param num_individuals: [int] Número de individuos que queremos generar.
        :param parameters: [ParameterVector] Parámetros del circuito generador de primeros padres.

        :return results_dict (dict): Diccionario con los valores generados para cada individuo.
        """

        # -- Creamos la lista de circuitos cuánticos
        _qc_list: List[QuantumCircuit] = []

        # -- Obtenemos la cantidad de parámetros
        _parameter_length: int = len([z for z in parameters])

        # -- Iteramos por cada individuo del total de individuos que se buscan generar
        for individual in range(0, num_individuals):

            # -- Iteramos por cada parámetro del diccionario de propiedades
            for parameter in range(0, _parameter_length):

                # -- Generamos el circuito cuantico de cada parametro
                _temp_qc: QuantumCircuit = self._generate_qc(num_qubits=self.max_qubits).copy()

                # -- Apendeamos el circuito cuántico generado a la lista de circuitos
                _qc_list.append(_temp_qc)

        # -- Ejecutamos todos los circuitos cuánticos bajo una misma sesion
        results = self.quantum_randomness(_qc_list)

        # -- Generamos un diccionario de resultados para adjudicar los parametros a cada individuo
        results_list: list = []

        # -- Llenamos el diccionario con los valores de _results
        for i, result in enumerate(results):
            results_list.append(self.binary_to_float_and_normalize_pi(result.keys()))

        # -- Retornamos los valores de los parámetros entre -pi y pi dividos por individuo
        # return self.dividir_lista(results_list, len([z for z in parameters]))
        return self.dividir_lista(results_list, 2)

    def generate_properties(self, vqc: QuantumCircuit, parameters: ParameterVector, vqc_parameters: List[list], num_qubits: int) -> np.ndarray:

        """
        Metodo para convertir los resultados de las VQC en valores normalizados de las propiedades
        :param vqc: Circuito cuántico variacional (VQC)
        :param parameters: Parámetros del circuito cuántico variacional (VQC)
        :param vqc_parameters: Parámetros con los que se instancias los parámetros del VQC
        :param num_qubits: Numero de qubits para obtener resultados
        :return: np.ndarray con los resultados normalizados de las propiedades
        """

        # -- Creamos todos los circuitos con sus respectivos parámetros
        _parameterized_circuits = []


        # -- Para conjunto de parámetros (en realidad -> ángulos de los parámetros)
        for circuit_parameters_values in vqc_parameters:
            _parameters_dict = {param: value for param, value in zip([z for z in parameters], circuit_parameters_values)}
            _parameterized_circuit = vqc.assign_parameters(_parameters_dict)
            _parameterized_circuits.append(_parameterized_circuit)

        # -- Ejecutamos el total de circuitos juntos en la misma sesión
        results: List[Dict[str, int]] = self.get_quantum_technology().run(_parameterized_circuits, shots=1024)

        # -- Extraemos resultados y los convertimos a propiedades
        all_properties: np.ndarray = np.zeros((len(vqc_parameters), num_qubits))

        for circuit_idx, counts in enumerate(results):
            for i in range(num_qubits):
                prob_one = sum(count for state, count in counts.items() if state[-(i + 1)] == '1')
                total = sum(counts.values())

                # -- Guardamos el resultado normalizado para el circuito correspondiente
                all_properties[circuit_idx, i] = prob_one / total if total > 0 else 0.5

        return all_properties

    @staticmethod
    def denormalize_hyperparameters(bounds_dict: Dict, all_hyperparameters: np.ndarray) -> np.ndarray:
        """
        Desnormaliza los hiperparámetros de acuerdo a los límites especificados en bounds_dict.
        :param bounds_dict: Diccionario de bounds del problema a resolver
        :param all_hyperparameters: Array de valores normalizados de los hiperparámetros.

        :return Matriz con los hiperparámetros desnormalizados.
        """

        # -- Convertimos las keys de bounds_dict a una lista para conservar el orden
        keys: list = list(bounds_dict.keys())

        # -- Inicializamos una matriz vacía para almacenar los resultados desnormalizados
        data_desnormalized = np.empty_like(all_hyperparameters)

        # -- Definimos variables en None
        min_val: int | float | None = None
        max_val: int | float | None = None

        # -- Para cada propiedad del diccionario
        for i, key in enumerate(keys):

            # -- Extraemos la información del diccionario
            bounds_info = bounds_dict[key]
            bound_type = bounds_info['bound_type']

            # -- Si el tipo de bound es interval obtenemos el valor mínimo y máximo
            if bound_type == "interval":
                min_val, max_val = bounds_info['limits']

            # -- Obtenemos los parámetros
            data_column = all_hyperparameters[:, i]

            # -- Desnormalizamos para un intervalo continuo
            if bound_type == 'interval':
                data_desnormalized[:, i] = data_column * (max_val - min_val) + min_val

                # -- Si el tipo de dato es 'int', lo convertimos a entero
                if bounds_info['type'] == 'int':
                    data_desnormalized[:, i] = np.round(data_desnormalized[:, i]).astype(int)

             # -- Desnormalizamos para un conjunto de valores predefinidos
            elif bound_type == 'predefined':
                predefined_values = np.array(bounds_info['limits'])

                # -- Aquí buscamos el valor más cercano en el arreglo de valores predefinidos
                indices = np.round(data_column * (len(predefined_values) - 1)).astype(int)
                data_desnormalized[:, i] = predefined_values[indices]

        return data_desnormalized

    @staticmethod
    def normalize_metric(metric_name, value, generations_fitness_statistics_df: pd.DataFrame):
        match metric_name:
            case 'accuracy' | 'recall' | 'specificity' | 'f1' | 'aur_roc' | 'precision' | 'negative_precision' | 'r2':
                # Normalizar directamente de [0, 1] a [-π, π]
                return -math.pi + 2 * math.pi * value
            case 'mae' | 'mse' :
                # Asumiendo que el valor está normalizado previamente a [0, 1]
                normalized_value = max(0, min(value, 1))  # Clamping entre 0 y 1
                return -math.pi + 2 * math.pi * normalized_value
            case 'other':
                # Si es "other", devolvemos un valor neutro o indicativo
                max_value = max(generations_fitness_statistics_df["max"].values)
                min_value = min(generations_fitness_statistics_df["min"].values)

                # -- Evita división por cero si todos los valores son iguales
                if max_value == min_value:
                    return 0.0

                # -- Normalizar value al rango [-π, π]
                normalized_value = -math.pi + 2 * math.pi * (value - min_value) / (max_value - min_value)
                return normalized_value

            case _:
                raise ValueError(f"Métrica desconocida: {metric_name}")

    @staticmethod
    def add_new_qubit(qc: QuantumCircuit, new_param: Parameter, bounds_dict: Dict, max_qubits: int) -> QuantumCircuit:

        """
        Añadimos nuevos qubits a un circuito cuántico.
        :param qc: [QuantumCircuit] Circuito cuántico al que queremos agregar qubits.
        :param new_param: [Parameter] Nuevo parámetros que queremos agregar al circuito.
        :param bounds_dict: [Dict] Diccionario de bounds con las propiedades y sus limites/malformaciones
        :param max_qubits: [int] Número máximo de qubits que se agregan al circuito // 2.
        """

        # -- Obtenemos el número existente de qubits
        num_existing_qubits: int = qc.num_qubits

        # -- Obtenemos las claves de bounds_dict == propiedades
        num_bounds_keys: int = len([z for z in bounds_dict.keys()])

        # -- Calculamos los qubits se deben agregar (división x 2 porque habrá 2 padres con misma cantidad de qubits)
        num_new_qubits = (max_qubits - num_bounds_keys) // 2

        # -- Creamos un nombre único para cada nuevo registro
        new_register = QuantumRegister(num_new_qubits, name="metrics")
        qc.add_register(new_register)

        # -- Aplicamos las rotaciones y CNOTs para cada nuevo qubit añadido
        for i in range(num_new_qubits):

            # -- La i representa el índice del nuevo qubit en el circuito
            qubit_index = num_existing_qubits + i

            if i == 0:
                division_factor = 1
            else:
                # El primero se divide por 2, el siguiente por 3, etc por dinamismo de propiedades
                division_factor = i + 1

            # -- Definimos los niveles de repetición escalada de las secciones del circuito que se repiten
            for deep_level in range(1, 5):

                # -- Añadimos las rotaciones en z y en y
                qc.ry(new_param / (deep_level * division_factor), qubit_index)

                if i % 2 == 0:
                    qc.ry(new_param / (deep_level * division_factor), qubit_index)
                else:
                    qc.rz(new_param / (deep_level * division_factor), qubit_index)
                # qc.rz(new_param / (deep_level * division_factor), qubit_index)

                # -- Entrelazamos con el último qubit anterior (si existe)
                if qubit_index > 0:
                    qc.cx(qubit_index - 1, qubit_index)

            # -- Conectamos con el primer qubit para cerrar el ciclo
            for idx in range(num_existing_qubits):
                qc.cx(qubit_index, idx)

        return qc

    @staticmethod
    def create_parent_vqc(bounds_dict: dict, vqc_1: QuantumCircuit, vqc_2: QuantumCircuit) -> QuantumCircuit:
        """
        Versión simplificada del circuito parental.
        Se enfoca en preservar diversidad y control sin entrelazamiento excesivo.
        """

        # Validación
        if vqc_1.num_qubits != vqc_2.num_qubits:
            raise ValueError("Ambos circuitos deben tener el mismo número de qubits.")

        num_qubits = vqc_1.num_qubits
        total_qubits = 2 * num_qubits
        num_props = len(bounds_dict)

        qr1 = QuantumRegister(num_qubits, "q1")
        qr2 = QuantumRegister(num_qubits, "q2")
        qc = QuantumCircuit(qr1, qr2)

        # Paso 1: Composición directa de padres
        qc.compose(vqc_1, qr1, inplace=True)
        qc.compose(vqc_2, qr2, inplace=True)
        qc.barrier(label="Padres directos")

        # Paso 2: Entrelazamiento simple 1:1 (mantener herencia directa)
        for i in range(num_props):
            qc.cx(qr1[i], qr2[i])
        qc.barrier(label="Entrelazamiento 1:1 básico")

        # Paso 3: Rotaciones suaves (introducir variación sin caos)
        angle = np.pi / 8
        for i in range(num_props):
            qc.ry(angle, qr1[i])
            qc.rz(angle, qr2[i])
        qc.barrier(label="Rotaciones suaves")

        # Paso 4: Toque de aleatoriedad mínima
        # H en qubits aleatorios solo si total_qubits > 4
        if total_qubits >= 4:
            aleatorios = random.sample(range(total_qubits), 2)
            for q in aleatorios:
                qc.h(q)
            qc.barrier(label="Aleatoriedad mínima")

        return qc

    def adding_measurements_parent_vqc(self, bounds_dict: dict, combined_qc: QuantumCircuit, verbose: bool, graph: bool) -> dict:

        """
        Añadimos las medidas a los circuitos cuánticos concatenados de los mejores padres
        :param bounds_dict: Diccionario con los bounds de las propiedades de los padres.
        :param combined_qc: Circuito cuántico concatenado (unión de los circuitos cuánticos de los mejores padres)
        circuitos por propiedad del bounds_dict
        :param verbose: Variable booleana que define si se pinta o no los circuitos en cada iteración
        :param graph: Variable booleana que define si se muestra el gráfico de circuitos con medidas.
        :return: [dict] Circuitos
        """

        # -- Determinamos el numero de qubits para las propiedades y el total de qubits por circuito
        num_property_qubits = len([z for z in bounds_dict.keys()])
        qubits_per_circuit = combined_qc.num_qubits // 2  # -- 2 porque tenemos dos circuitos cuánticos concatenados

        circuits: dict = {}

        # -- Iteramos por las propiedades de los bounds
        for pid, parameter in enumerate([z for z in bounds_dict.keys()]):

            # -- Creamos una copia del circuito combinado
            combined_qc_duplicated = combined_qc.copy()

            # -- Agregamos registros clásicos para guardar los resultados de las mediciones
            qc_meas: QuantumCircuit = QuantumCircuit(combined_qc_duplicated.num_qubits)
            qc_meas.compose(combined_qc_duplicated, inplace=True)

            # -- Creamos y agregamos un registro clásico para las mediciones
            cr: ClassicalRegister = ClassicalRegister(combined_qc_duplicated.num_qubits, name='c')
            qc_meas.add_register(cr)

            # -- Definimos qué qubit se va a medir por parámetro (agregamos qubits de propiedades -ambos circuitos-)
            qubits_to_measure: list = [pid, qubits_per_circuit + pid]

            # -- Agregamos el resto de qubits a medir del primer circuito (los qubits de las métricas)
            for i in range(num_property_qubits, qubits_per_circuit):
                qubits_to_measure.append(i)

            # -- Agregamos el resto de qubits a medir del segundo circuito (los qubits de las métricas)
            for i in range(qubits_per_circuit + num_property_qubits, 2 * qubits_per_circuit):
                qubits_to_measure.append(i)

            # -- Agregamos las medidas propiamente dicho para todos los qubits seleccionados
            for qubit_idx in qubits_to_measure:
                qc_meas.measure(qubit_idx, qubit_idx)

            circuits[parameter] = qc_meas

            if verbose and graph is True:
                self.advanced_qc_plot(qc_meas)

        return circuits

    @staticmethod
    def advanced_qc_plot(combined_qc: QuantumCircuit):
        """
        Metodo avanzado de graficación de circuitos cuánticos
        :param combined_qc: Circuito cuántico a plotear
        :return:
        """

        # -- Dibujamos el circuito
        fig, ax = plt.subplots(figsize=(20, 10))

        # -- Customizamos parámetros del gráfico
        combined_qc.draw(
            "mpl",
            fold=-1,  # -- Previene folding
            scale=0.8,  # -- Escalamos los elementos
            style={
                "fontsize": 48,  # -- 48 Fuente primaria
                "subfontsize": 36,  # -- 36 Fuente secundaria
                "circuit_label_margin": 0.1,  # -- Ajuste de márgenes
                "gate_offset": 0.2,  # -- Espacio entre puertas
                "showindex": True,  # -- Muestra el índice de cada qubit
                "figwidth": 20,  # -- Ancho de la figura en inches
                "dpi": 4096,  # -- Alta resolución
                "cregbundle": False  # -- No aprieta los registros clásicos
            },
            ax=ax
        )

        plt.tight_layout()
        plt.show()

    @staticmethod
    def _generate_qc(num_qubits=32) -> QuantumCircuit:
        """
        Genera un circuito cuántico que genera números aleatorios dentro de un rango especificado por bounds_dict.
        :param num_qubits: Número de qubits que se utilizan para generar un valor que luego será denormalizado (-pi/pi)

        Returns: QuantumCircuit: Circuito cuántico listo para ser ejecutado.
        """

        # -- Creamos el circuito cuántico con qubits y bits clásicos
        qc = QuantumCircuit(num_qubits, num_qubits)

        # -- Aplicamos puertas Hadamard a todos los qubits para crear superposición
        qc.h(range(num_qubits))

        # -- Añadimos medidas a todos los qubits
        qc.measure(range(num_qubits), range(num_qubits))

        return qc

    @staticmethod
    def binary_to_float_and_normalize_pi(binary_str):

        # -- Convertimos la cadena binaria a un entero
        int_value = int([z for z in binary_str][0], 2)

        # -- Normalizamos el entero a un valor entre 0 y 1
        normalized_value = int_value / (2 ** len([z for z in binary_str][0]) - 1)

        # -- Mapeamos el valor normalizado al rango [-pi, pi]
        mapped_value = -np.pi + 2 * np.pi * normalized_value

        # -- Redondeamos a 10 decimales
        return round(mapped_value, 10)

    @staticmethod
    def rescaling_pi_to_integer_float(bounds_dict: dict, parameter: str, pi_number: float):
        """
        Reescala un valor desde el rango [-π, π] al rango especificado en bounds_dict.

        Args:
            bounds_dict: Diccionario con los límites de cada parámetro
            parameter: Nombre del parámetro a reescalar
            pi_number: Valor en el rango [-π, π] que se desea reescalar

        Returns:
            El valor reescalado, como entero si los límites son enteros,
            o como flotante con 10 decimales si los límites son flotantes.
        """
        # Obtener los límites del parámetro
        min_bound = min(bounds_dict[parameter]["malformation_limits"])
        max_bound = max(bounds_dict[parameter]["malformation_limits"])

        # Rango actual de pi_number (-pi, pi)
        current_min, current_max = -np.pi, np.pi

        # Realizar el mapeo lineal desde [-pi, pi] al rango deseado
        mapped_value = min_bound + (pi_number - current_min) * (max_bound - min_bound) / (current_max - current_min)

        # Si los límites son enteros, redondear a entero. Si no, redondear a 10 decimales.
        if isinstance(min_bound, int) and isinstance(max_bound, int):
            return int(round(mapped_value))
        else:
            return round(mapped_value, 10)

    @staticmethod
    def dividir_lista(lista: List[float], tamanio_grupo: int) -> List[List[float]]:
        return [lista[i:i + tamanio_grupo] for i in range(0, len(lista), tamanio_grupo)]

    def quantum_randomness(self, qcs: List[QuantumCircuit]):
        """
        Ejecuta circuitos cuánticos para generar números aleatorios utilizando un simulador u ordenador cuántico real.
        :param qcs: (List[QuantumCircuit]) Lista de circuitos cuánticos a ejecutar.

        Return: results: (List[Dict[str: int]) Lista de resultados en formato binario.
        """

        # -- Instanciamos el ejecutor del circuito
        _executor = self.get_quantum_technology()

        # -- Ejecutamos una vez el circuito cuántico con el ejecutor
        result: list = _executor.run(qcs, 1)

        return result

    def get_permutations_hybrid(self, bounds_dict: tuple, log2: int, shots: int) -> List:
        """
        Metodo híbrido para generar individuos para problema de tipo totally_restricted.
        Se utiliza un circuito cuántico combinado con procesamiento clásico.

        :param bounds_dict: [Dict] Diccionario de bounds.
        :param log2: [int] Número de permutaciones a generar.
        :param shots: Número de ciudades (puede ser grande, como 10).

        :return: [List] Lista de permutaciones generadas (potencial individuo)
        """

        # -- Definimos el límite minimo y máximo de los valores que se pueden generar
        min_val = min(bounds_dict["limits"])
        max_val = max(bounds_dict["limits"])
        value_range = max_val - min_val + 1

        # -- Utilizamos un circuito cuántico para generar "semillas" aleatorias
        qc = QuantumCircuit(log2, log2)

        # -- Añadimos puertas Hadamard a todos los qubits
        for i in range(log2):
            qc.h(i)

        # -- Añadimos mediciones a todos los qubits
        qc.measure(range(log2), range(log2))

        # -- Ejecutamos el circuito y obtenemos sus resultados
        results: List[Dict[str, int]] = self.execution_object.run([qc], shots=shots)

        # -- Obtenemos la primera semilla válida del resultado
        bitstring = next(iter(results[0].keys()))
        seed = int(bitstring, 2)
        random.seed(seed)

        # -- Generamos la permutación aleatoria (un solo individuo)
        individual = list(range(value_range))
        random.shuffle(individual)

        return individual

    # ---------------------------- Metodo específicos del QAOA -------------------------------------- #
    @staticmethod
    def create_tsp_qubo(element_matrix: Dict,
                        problem_type: Literal['minimize', 'maximize'],
                        return_to_origin: Literal['return_to_origin', 'no_return'] | None,
                        contains_zero: bool) -> QuadraticProgram:
        """
        Convierte un problema TSP a un QUBO.
        Si el nodo 0 está en el clúster, se fuerza como origen.

        :param element_matrix: [Dict] Diccionario {ID: (x, y)} con nodos del TSP.
        :param problem_type: [Literal['minimize', 'maximize']] Define el tipo de problema si es de minimización o maximización.
        :param return_to_origin: [Literal['return_to_origin', 'no_return'] | None] Variable que define si el problema retorna al origen o no.
        :param contains_zero: Define si el clúster posee el valor 0 (ej. Origen del recorrido).
        :return: Problema QUBO generado.
        """

        # -- Mapeamos índices internos para trabajar más fácil en matrices
        ids = list(element_matrix.keys())
        id_to_index = {node_id: idx for idx, node_id in enumerate(ids)}

        # -- Obtenemos la cantidad de elementos del clúster
        n = len(ids)

        # -- Matriz de distancias
        dist = np.zeros((n, n))
        for i in range(n):
            for j in range(n):
                if i != j:
                    dist[i, j] = np.linalg.norm(
                        np.array(element_matrix[ids[i]]) - np.array(element_matrix[ids[j]])
                    )

        qp = QuadraticProgram("TSP")

        # -- Declaramos variables binarias x_{i}_{j}
        for i in range(n):
            for j in range(n):
                if i != j:
                    qp.binary_var(f'x_{i}_{j}')

        # -- Función objetivo: minimizamos o maximizamos la suma de valores (ej. Las distancias entre ciudades)
        linear_obj = {f'x_{i}_{j}': dist[i, j] for i in range(n) for j in range(n) if i != j}
        qp.minimize(linear=linear_obj) if problem_type == "minimize" else qp.maximize(linear=linear_obj)

        # -- Si el cluster contiene el nodo 0 y no debe retornar al origen...
        if return_to_origin == "no_return" and contains_zero:

            # -- Obtenemos el índice del elemento 0
            zero_index = id_to_index[0]

            # -- Forzamos la salida desde el nodo 0
            qp.linear_constraint({f'x_{zero_index}_{j}': 1 for j in range(n) if j != zero_index}, '==', 1)

            # -- Prohibimos el regreso al nodo 0
            for i in range(n):
                if i != zero_index:
                    # Fijamos directamente la variable a 0
                    qp.linear_constraint({f'x_{i}_{zero_index}': 1}, '==', 0)

            # -- Hacemos que el resto de los nodos tengan entrada y salida una vez
            for j in range(n):
                if j != zero_index:
                    qp.linear_constraint({f'x_{i}_{j}': 1 for i in range(n) if i != j}, '==', 1)

            for i in range(n):
                if i != zero_index:
                    qp.linear_constraint({f'x_{i}_{j}': 1 for j in range(n) if j != i}, '==', 1)

        else:
            # -- En el resto de casos... todos los nodos entran y salen una sola vez
            for j in range(n):
                qp.linear_constraint({f'x_{i}_{j}': 1 for i in range(n) if i != j}, '==', 1)

            for i in range(n):
                qp.linear_constraint({f'x_{i}_{j}': 1 for j in range(n) if j != i}, '==', 1)

        # -- Convertimos el problema a formato QUBO
        qubo_converter = QuadraticProgramToQubo()
        return qubo_converter.convert(qp)

    @staticmethod
    def decode_solution(x: Dict, n: int, contains_zero: bool = False):
        """
        Decodifica la solución del QUBO.
        - Evita ciclos (rutas cerradas), incluyendo volver al nodo inicial.
        - Si contains_zero es True, 0 se pone al inicio.
        - Completa nodos faltantes aleatoriamente.

        :param x: [Dict] Diccionario con claves tipo 'x_i_j' y valores 0/1.
        :param n: Número total de nodos.
        :param contains_zero: Si el nodo 0 debe estar al inicio.
        :return: Lista con el orden de los nodos.
        """

        # -- Creamos el diccionario de nodos: de -> a
        edges = {int(k.split('_')[1]): int(k.split('_')[2]) for k, v in x.items() if v > 0.5 and k.startswith('x_')}

        # -- Creamos un set para evitar duplicados y la lista de combinaciones
        repeated = set()
        combination = []

        # -- Elegimos el nodo inicial (si el clúster posee 0 no aseguramos que es el punto de partida)
        start = 0 if contains_zero else next(iter(edges), 0)
        current = start
        combination.append(current)
        repeated.add(current)

        # -- Construimos la combinación evitando ciclos
        while current in edges:
            next_node = edges[current]

            # -- Si ya fue utilizado, cortamos
            if next_node in repeated:
                break

            combination.append(next_node)
            repeated.add(next_node)
            current = next_node

        # -- Si el último nodo apunta al primero, eliminamos el cierre
        if len(combination) > 1 and edges.get(combination[-1], None) == combination[0]:
            del edges[combination[-1]]

        # -- Agregamos nodos faltantes aleatoriamente
        missing = [i for i in range(n) if i not in repeated]
        random.shuffle(missing)
        combination.extend(missing)

        # -- Si contains_zero, aseguramos que 0 esté al inicio
        if contains_zero and combination[0] != 0:
            if 0 in combination:
                combination.remove(0)
            combination.insert(0, 0)

        return combination

    def solve_tsp_with_qaoa(self, element_matrix: Dict, problem_type: Literal['minimize', 'maximize'], contains_zero: bool, return_to_origin: Literal['return_to_origin', 'no_return'] | None):
        """
        Metodo para resolver el problema de optimización combinatoria mediante un QAOA.
        :param element_matrix: [Dict] Matriz de elementos del individuo que se deben combinar de la mejor manera.
        :param problem_type: Literal['minimize', 'maximize'] Definimos si el problema es de minimización o maximización.
        :param contains_zero: [bool] Variable booleana para saber si el clúster empieza con 0.
        :param return_to_origin: [Literal['return_to_origin', 'no_return'] | None Variable booleana para saber si el problema debe volver al origen.

        :return: combination, total_value, element_indices
        """

        # -- Calculamos la cantidad de elementos en la matriz de elementos (elementos del clúster)
        n = len([z for z in element_matrix.keys()])


        # -- Ejecutamos que transforma el problema a resolver en un problema de tipo QUBO
        qubo: QuadraticProgram = self.create_tsp_qubo(element_matrix=element_matrix,
                                                      problem_type=problem_type,
                                                      return_to_origin=return_to_origin,
                                                      contains_zero=contains_zero)

        # -- TODO: Definir si es simulador u ordenador cuántico

        # -- Definimos la tecnología cuántica de ejecución del QAOA
        sampler = AerSampler()

        # -- Definimos el optimizador del QAOA
        optimizer = COBYLA(maxiter=1000)

        # -- Instanciamos el QAOA
        qaoa = QAOA(sampler=sampler, optimizer=optimizer, reps=5)
        qaoa_optimizer = MinimumEigenOptimizer(qaoa)

        # -- Resolvemos el problema qubo con el optimizador
        result = qaoa_optimizer.solve(qubo)

        # -- Obtenemos el resultado para cada variable
        x = result.variables_dict

        # -- Decodificamos la solución
        element_indices = self.decode_solution(x, n, contains_zero=contains_zero)

        # -- Si no retornamos indices de elementos se devuelve una lista vacía
        if len(element_indices) < 1:
            return [], float("inf"), []

        # -- Obtenemos los ids de los elementos
        element_ids = list(element_matrix.keys())

        # -- Remapeamos la combinación a los elementos originales
        combination = [element_ids[i] for i in element_indices]

        return combination

    # ---------------------------- Metodos específicos para TSP  ------------------------------------- #
    @staticmethod
    def cluster_distances(element_matrix: Dict[str, tuple], num_clusters: int):
        """
        Metodo para separar los elementos del individuo en clústers.
        :param element_matrix: Matriz con valores de los elementos del individuo (ej: distancias entre ciudades -TSP-).
        :param num_clusters: Número de clústers que se buscan generar.
        :return: Clústers y centroides de cada clúster.
        """

        # -- Obtenemos los valores de la matriz de elementos
        coords = np.array(list(element_matrix.values()))

        # -- Instanciamos el KMeans con el número de clústers pretendido
        kmeans = KMeans(n_clusters=num_clusters)

        # -- Ejecutamos el modelo no-supervisado de clústering
        labels = kmeans.fit_predict(coords)

        # -- Generamos los clísters
        clusters = {i: {} for i in range(num_clusters)}
        for distance_id, label in zip(element_matrix.keys(), labels):
            clusters[label][distance_id] = element_matrix[distance_id]

        return clusters, kmeans.cluster_centers_

    @staticmethod
    def plot__restricted_solution(cities, order, title="Ruta completa"):
        plt.figure(figsize=(8, 5))
        coords = [cities[i] for i in order]
        x, y = zip(*coords)
        plt.plot(x, y, 'o-b')
        for i, city in enumerate(order[:-1]):
            cx, cy = cities[city]
            plt.text(cx + 0.01, cy + 0.01, str(city), fontsize=9)
        plt.title(title)
        plt.grid(True)
        plt.axis("equal")
        plt.show()

    @staticmethod
    def euclidean(p1: tuple, p2: tuple) -> float:
        """
        Metodo para medir distancias eucledianas
        :param p1: Conjunto de coordenadas del individuo 1
        :param p2: Conjunto de coordenadas del individuo 2
        :return: Distancias eucledianas
        """
        return math.sqrt((p1[0] - p2[0]) ** 2 + (p1[1] - p2[1]) ** 2)

__all__ = ['QuantumSimulator', 'QuantumMachine', 'QuantumTechnology']



from typing import Iterable, Literal


class BoundCreator:
    def __init__(self):
        """
        Clase para crear el diccionario de bounds y de restricciones
        """
        self.bound_object: dict = {}

    def add_interval_bound(self, parameter: str,
                           parameter_low_limit: int | float,
                           parameter_high_limit: int | float,
                           malformation_low_limit: int | float | None,
                           malformation_high_limit: int | float | None,
                           parameter_type: Literal['int', 'float'] = "int",
                           ):
        """
        Metodo para crear bounds que se pasarán al algoritmo genético cuántico
        :param parameter: nombre del parametro, ej 'learning_rate'
        :param parameter_low_limit: Limite inferior que quieres que pueda tener el parametro, ej, 0.00001
        :param parameter_high_limit: limite superior que quieres que tenga el parametro, ej: 0.1
        :param malformation_low_limit: Limite inferior a partir del cual se considerará que el individuo tiene una malformacion
        :param malformation_high_limit: Limite superior a partir del cual se considerará que el individuo tiene una malformacion
        :param parameter_type: Elegir entre ['int', 'float']. int para parametros que no pueden tener valores continus, como el batch_size, float para los que sí como learning rate
        """
        self.bound_object = self.bound_object | IntervalBound(parameter, parameter_low_limit, parameter_high_limit, malformation_low_limit, malformation_high_limit, parameter_type)

    def add_predefined_bound(self, parameter: str,
                             parameter_list: Iterable,
                             parameter_type: Literal['int', 'float'] = "int",
                             ):
        """
        Método para crear bounds que se pasarán al algoritmo genético cuántico
        :param parameter: nombre del parametro, ej 'learning_rate'
        :param parameter_list: tupla que va a contener los posibles valores para el parámetro
        :param parameter_type: Elegir entre ['int', 'float']. int para parametros que no pueden tener valores continus, como el batch_size, float para los que sí como learning rate
        """
        self.bound_object = self.bound_object | PredefinedBound(parameter, parameter_list, parameter_type)

    def get_bound(self):
        return self.bound_object


class IntervalBound(dict):
    def __init__(self,
                 parameter: str,
                 parameter_low_limit: int | float,
                 parameter_high_limit: int | float,
                 malformation_low_limit: int | float | None,
                 malformation_high_limit: int | float | None,
                 parameter_type: Literal['int', 'float'] = "int"):
        """
        Clase para crear bounds que se pasarán al algoritmo genético cuántico
        :param parameter: nombre del parametro, ej 'learning_rate'
        :param parameter_low_limit: Limite inferior que quieres que pueda tener el parametro, ej, 0.00001
        :param parameter_high_limit: limite superior que quieres que tenga el parametro, ej: 0.1
        :param malformation_low_limit: Limite inferior a partir del cual se considerará que el individuo tiene una malformacion
        :param malformation_high_limit: Limite superior a partir del cual se considerará que el individuo tiene una malformacion
        :param parameter_type: Elegir entre ['int', 'float']. int para parametros que no pueden tener valores continus, como el batch_size, float para los que sí como learning rate
        """

        if parameter_type not in ['int', 'float']:
            raise ValueError(f"Bound: Error al crear el Bound, el parameter type solo puede contener estos valores: ['int', 'float']. Valor asignado: {parameter_type}")

        bound_data = {
            parameter: {
                "limits": (parameter_low_limit, parameter_high_limit),
                "type": parameter_type,
                "malformation_limits": (malformation_low_limit, malformation_high_limit),
                "bound_type": "interval"
            }
        }

        super().__init__(bound_data)


class PredefinedBound(dict):
    def __init__(self,
                 parameter: str,
                 parameter_list: Iterable,
                 parameter_type: Literal['int', 'float'] = "int"):
        """
        Clase para crear bounds que se pasarán al algoritmo genético
        :param parameter: nombre del parametro, ej 'learning_rate'
        :param parameter_type: Elegir entre ['int', 'float']. int para parametros que no pueden tener valores continus, como el batch_size, float para los que sí como learning rate
        """

        if parameter_type not in ['int', 'float']:
            raise ValueError(f"Bound: Error al crear el Bound, el parameter type solo puede contener estos valores: ['int', 'float']. Valor asignado: {parameter_type}")

        bound_data = {
            parameter: {
                "limits": parameter_list,
                "type": parameter_type,
                "malformation_limits": parameter_list,
                "bound_type": "predefined"
            }
        }

        super().__init__(bound_data)

__all__ = ['BoundCreator', 'IntervalBound', 'PredefinedBound']



__version__ = "0.1.12"

# Importar las clases
from aletheia_quantum_genetic_optimizers.individuals import Individual
from aletheia_quantum_genetic_optimizers.bounds import BoundCreator
from aletheia_quantum_genetic_optimizers.genetic_optimizer import GenethicOptimizer

# Exportar las clases para que sean accesibles directamente desde el paquete
__all__ = ['Individual', 'BoundCreator', 'GenethicOptimizer']



from abc import ABC, abstractmethod
from aletheia_quantum_genetic_optimizers.individuals import Individual
from typing import List, Literal
import random
from info_tools import InfoTools


class Tournament(ABC):
    def __init__(self, podium_size: int = 3, problem_type: Literal["minimize", "maximize"] = "minimize", verbose: bool = False):
        self.podium_size: int = podium_size
        self.problem_type: Literal["minimize", "maximize"] = problem_type
        self.verbose: bool = verbose
        self.IT: InfoTools = InfoTools()

    @abstractmethod
    def run_tournament(self, individuals_list: List[Individual]):
        pass


class EaSimple(Tournament):
    def __init__(self, podium_size: int = 3, problem_type: Literal["minimize", "maximize"] = "minimize", verbose: bool = False):
        super().__init__(podium_size, problem_type, verbose)

    def run_tournament(self, individuals_list: List[Individual]) -> List[Individual]:

        """
        Metodo de torneo EaSimple
        :param individuals_list: Lista de individuos de una generación

        :return: La lista de los individuos ganadores
        """

        # -- Filtramos los individuos con malformaciones
        individuals_list: list = [ind for ind in individuals_list if not ind.get_individual_malformation()]

        # -- Definimos la lista de ganadores del torneo
        selected_individuals = []

        while individuals_list:

            # -- Seleccionamos hasta podium_size individuos aleatoriamente
            competitors = random.sample(individuals_list, min(self.podium_size, len(individuals_list)))

            # -- Seleccionamos el mejor según el tipo de problema
            winner = max(competitors, key=lambda ind: ind.get_individual_fitness()) if self.problem_type == "maximize" else min(competitors, key=lambda ind: ind.get_individual_fitness())
            selected_individuals.append(winner)

            # -- Eliminamos todos los competidores de la lista (no solo el ganador)
            individuals_list = [ind for ind in individuals_list if ind not in competitors]

        return selected_individuals


__all__ = ['Tournament', 'EaSimple']





from aletheia_quantum_genetic_optimizers.variability_explossion_methods import CrazyVariabilityExplossion
from aletheia_quantum_genetic_optimizers.reproduction_methods import Reproduction
from aletheia_quantum_genetic_optimizers.population_methods import Population
from aletheia_quantum_genetic_optimizers.mutation_methods import Mutation
from aletheia_quantum_genetic_optimizers.individuals import Individual
from aletheia_quantum_genetic_optimizers.tournament_methods import *

from typing import Callable, Dict, Literal, List
from info_tools import InfoTools

import matplotlib.pyplot as plt
import pandas as pd
import numpy as np


class GenethicOptimizer:
    def __init__(self,
                 bounds_dict: Dict,
                 num_generations: int,
                 num_individuals: int,
                 max_qubits: int,
                 objective_function: Callable,
                 metric_to_optimize: Literal['accuracy', 'recall', 'specificity', 'f1', 'aur_roc', 'precision', 'negative_precision', 'mae', 'mse', 'other', 'r2'],
                 problem_restrictions: Literal['bound_restricted', 'totally_restricted'] = "bound_restricted",
                 return_to_origin: Literal['return_to_origin', 'no_return'] | None = None,
                 problem_type: Literal["minimize", "maximize"] = "minimize",
                 tournament_method: Literal["ea_simple"] = "ea_simple",
                 podium_size: int = 3,
                 mutate_probability: float = 0.25,
                 mutate_gen_probability: float = 0.2,
                 mutation_policy: Literal['soft', 'normal', 'hard'] = 'normal',
                 verbose: bool = True,
                 early_stopping_generations: Literal['gradient'] | int = 'gradient',
                 variability_explossion_mode: Literal['crazy'] = 'crazy',
                 variability_round_decimals: int = 3,
                 randomness_quantum_technology: Literal["simulator", "quantum_machine"] = "simulator",
                 randomness_service: Literal["aer", "ibm"] = "aer",
                 optimization_quantum_technology: Literal["simulator", "quantum_machine"] = "simulator",
                 optimization_service: Literal["aer", "ibm"] = "aer",
                 qm_api_key: str | None = None,
                 qm_connection_service: Literal["ibm_quantum", "ibm_cloud"] | None = None,
                 quantum_machine: Literal["ibm_brisbane", "ibm_kyiv", "ibm_sherbrooke", "least_busy"] = "least_busy",
                 element_matrix: Dict[str, tuple] | None = None
                 ):
        """
        Clase principal para crear un algoritmo genético cuántico con el objetivo de resolver problemas de optimización
        combinatoria con distintos valores entre bounds y problemas de optimización combinatoria con valores concretos.

        :param bounds_dict: Diccionario en el que se definen los parámetros a optimizar y sus valores, ej. '{learning_rate: (0.0001, 0.1)}'
        :param num_generations: Numero de generaciones que se van a ejecutar
        :param num_individuals: Numero de Individuos iniciales que se van a generar
        :param max_qubits: Numero máximo de qubits a emplear para reproducir individuos (define el numero entero y la parte decimal de los numeros enteros y flotantes que se quieren generar)
        :param objective_function: Función objetivo que se va a emplear para puntuar a cada individuo (debe retornar un float)
        :param metric_to_optimize: Metrica que se quiere optimizar ['accuracy', 'recall', 'specificity', 'f1',
        'aur_roc', 'precision', 'negative_precision', 'mae', 'mse', 'other'] -> other significa cualquier otra genérica.
        Por ejemplo, se puede utilizar other para un problema de optimización de tipo viajante de comercio.
        :param problem_restrictions: ['bound_restricted', 'totally_restricted'] Restricciones que se van a aplicar a la hora de crear individuos, reprocirlos y mutarlos
        :param return_to_origin: [Literal['return_to_origin', 'no_return'] | None] En caso de problemas totally_restricted es necesario saber si el problema termina en el origen o no es necesario que suceda esto
        :param problem_type: [minimize, maximize] Seleccionar si se quiere minimizar o maximizar el resultado de la función objetivo. Por ejemplo si usamos un MAE es minimizar,
         un Accuracy sería maximizar.
        :param tournament_method: [easimple, .....] Elegir el tipo de torneo para seleccionar los individuos que se van a reproducir.
        :param podium_size: Cantidad de individuos de la muestra que van a competir para elegir al mejor. Por ejemplo, si el valor es 3, se escogen iterativamente 3 individuos
        al azar y se selecciona al mejor. Este proceso finaliza cuando ya no quedan más individuos y todos han sido seleccionados o deshechados.
        :param mutate_probability:Tambien conocido como indpb ∈[0, 1]. Probabilidad de mutar que tiene cada gen. Una probabilidad de 0, implica que nunca hay mutación,
        una probabilidad de 1 implica que siempre hay mutacion.
        :param mutate_gen_probability: [float] Probabilidad de mute un gen
        :param mutation_policy: Literal['soft', 'normal', 'hard'] Política de mutación (liviana, estandar y agresiva),
        :param verbose: Variable que define si se pinta información extra en consola y si se generan los graficos de los circuitos cuánticos.
        :param early_stopping_generations: Cantidad de generaciones que van a transcurrir para que en caso de repetirse el min o max del fitness, se active el modo variability_explosion
        :param variability_explossion_mode: Modo de explosion de variabilidad, es decir, que se va a hacer para intentar salir de un minimo local establecido
        :param variability_round_decimals: Decimales a los que redondear las estadisticas de cálculo del min o max para la explosion de variabilidad. Por ejemplo,
        en un caso de uso que busque accuracy, podría ser con 2 o 3 decimales. para casos de uso que contengan números muy bajos, habría que agregar más.
        :param randomness_quantum_technology. Literal["simulator", "quantum_machine"]. Tecnología cuántica con la que calculan la lógica. Si es simulator, se hará con un simulador definido en el
        parámetro service. Si es quantum_machine, el algoritmo se ejecutará en una máquina cuántica definida en el parámetro technology.  Se aplica al cálculo de numeros random (primera generación).
        :param randomness_service: ["aer", "ibm"] El servicio tecnológico con el cual se ejecuta la lógica. Se aplica al cálculo de numeros random (primera generación).
        :param optimization_quantum_technology. Literal["simulator", "quantum_machine"]. Tecnología cuántica con la que calculan la lógica. Si es simulator, se hará con un simulador definido en el
        parámetro service. Si es quantum_machine, el algoritmo se ejecutará en una máquina cuántica definida en el parámetro technology. Se aplica al proceso de optimización.
        :param optimization_service. ["aer", "ibm"] El servicio tecnológico con el cual se ejecuta la lógica. Se aplica al proceso de optimización.
        :param qm_api_key. API KEY para conectarse con el servicio de computación cuántica de una empresa.
        :param qm_connection_service. Literal["ibm_quantum", "ibm_cloud"] | None. Servicio específico de computación cuántica. Por ejemplo, en el caso de IBM pueden ser a la fecha ibm_quantum | ibm_cloud
        :param quantum_machine. Literal["ibm_brisbane", "ibm_kyiv", "ibm_sherbrooke", "least_busy"]. Nombre del ordenador cuántico a utilizar. Por ejemplo, en el caso de IBM puede ser ibm_brisbane, ibm_kyiv, ibm_sherbrooke. Si se deja en least_busy,
        se buscará el ordenador menos ocupado para llevar a cabo la ejecución del algoritmo cuántico.
        :param element_matrix: Dict[str, tuple] | None = Matriz de distancia utilizada para los problemas de optimización combinatoria de tipo TSP.
        """

        # -- Almacenamos propiedades
        self.bounds_dict: Dict = bounds_dict
        self.num_generations: int = num_generations
        self.num_individuals: int = num_individuals
        self.max_qubits: int = max_qubits
        self.objective_function: Callable = objective_function
        self.metric_to_optimize: Literal['accuracy', 'recall', 'specificity', 'f1', 'aur_roc', 'precision', 'negative_precision', 'mae', 'mse', 'other', 'r2'] = metric_to_optimize
        self.problem_restrictions: Literal['bound_restricted', 'totally_restricted'] = problem_restrictions
        self.return_to_origin: Literal['return_to_origin', 'no_return'] | None = return_to_origin
        self.problem_type: Literal["minimize", "maximize"] = problem_type
        self.tournament_method: str = tournament_method
        self.podium_size: int = podium_size
        self.mutate_probability: float = mutate_probability
        self.mutate_gen_probability: float = mutate_gen_probability
        self.mutation_policy: Literal['soft', 'normal', 'hard'] = mutation_policy
        self.verbose: bool = verbose
        self.early_stopping_generations: int = early_stopping_generations if isinstance(early_stopping_generations, int) else max(int(self.num_generations * 0.05), 3)
        self.early_stopping_generations_executed: bool = False
        self.early_stopping_generations_executed_counter: int = 0
        self.variability_round_decimals: int = variability_round_decimals
        self.randomness_quantum_technology: Literal["simulator", "quantum_machine"] = randomness_quantum_technology
        self.randomness_service: Literal["aer", "ibm"] = randomness_service
        self.optimization_quantum_technology: Literal["simulator", "quantum_machine"] = optimization_quantum_technology
        self.optimization_service: Literal["aer", "ibm"] = optimization_service
        self.qm_api_key: str | None = qm_api_key
        self.qm_connection_service: Literal["ibm_quantum", "ibm_cloud"] | None = qm_connection_service
        self.quantum_machine: Literal["ibm_brisbane", "ibm_kyiv", "ibm_sherbrooke", "least_busy"] = quantum_machine
        self.element_matrix: Dict[str, tuple] | None = element_matrix

        # -- MEJOR INDIVIDUAL
        self.best_individual_values: dict | None = None

        # <editor-fold desc="DEFINICIÓN DE VARIABLES GENERALES DEL ALGORITMO GENÉTICO CUÁNTICO">

        # -- Instanciamos info tools para los prints
        self.IT: InfoTools = InfoTools()

        self.IT.header_print("Se definen las variables generales del AletheIA Quantum Genetic Optimizer")

        # -- Definimos la variable que controla la variabilidad de los resultados de la función objetivo por generación
        self.variability_explosion_starts_in_generation: int | None = None

        # -- Instanciamos la clase GenethicTournamentMethods (clase que define el torno y lo almacenamos)
        self.IT.sub_intro_print("Instanciando el tipo de torneo...")
        self.GTM: Tournament = self.get_tournament_method(self.verbose)

        # -- Instanciamos la clase que controla que la variabilidad (sin cambios --> aplicamos mutación agresiva)
        self.IT.sub_intro_print("Instanciando la monitorización de variabilidad de resultados...")

        match variability_explossion_mode:
            case 'crazy':
                self.VEM: CrazyVariabilityExplossion = CrazyVariabilityExplossion(self.early_stopping_generations,
                                                                                  self.problem_type,
                                                                                  self.variability_round_decimals,
                                                                                  self.verbose)
            case _:
                self.VEM: CrazyVariabilityExplossion = CrazyVariabilityExplossion(self.early_stopping_generations,
                                                                                  self.problem_type,
                                                                                  self.variability_round_decimals,
                                                                                  self.verbose)

        self.IT.info_print("Instanciación de la monitorización de variabilidad de resultados realizada con éxito")

        # -- Validamos los inputs
        self.IT.sub_intro_print("Validando los inputs...")
        self.validate_input_parameters()
        self.IT.info_print("Validación de los inputs realizada con éxito")

        if self.verbose:
            self.IT.header_print(f"Bounds_dict y valores a combinar")
            for k, v in self.bounds_dict.items():
                self.IT.info_print(f"{k}: {v}")

        # </editor-fold>

        # <editor-fold desc="CREACIÓN Y EVALUACIÓN DE LA FUNCIÓN DE COSTE DE LA GENERACIÓN 0">

        self.IT.header_print("Creamos la población 0")

        # -- Creamos el objeto poblacion y la poblacion inicial
        self.population: population = Population(self.bounds_dict,
                                                 self.num_individuals,
                                                 self.problem_restrictions,
                                                 self.variability_round_decimals)

        # -- Creamos las listas de individuos que vamos a ir usando
        self.population.create_population(quantum_technology=self.randomness_quantum_technology,
                                          service=self.randomness_service,
                                          qm_api_key=self.qm_api_key,
                                          qm_connection_service=self.qm_connection_service,
                                          quantum_machine=self.quantum_machine,
                                          max_qubits=self.max_qubits,
                                          element_matrix=self.element_matrix)

        if self.verbose:

            self.IT.sub_intro_print("Individuos de la generación 0")

            for individual in self.population.populuation_dict[str(0)]:

                match self.problem_restrictions:

                    case "bound_restricted":
                       self.IT.info_print(f"INDIVIDUO: {individual.get_individual_id()} --> PROPIEDADES: {individual.get_individual_values()} --> PARAMETROS: {individual.get_vqc_parameters_values()}")

                    case "totally_restricted":
                        self.IT.info_print(f"INDIVIDUO: {individual.get_individual_id()} --> PROPIEDADES: {individual.get_individual_values()}")

        # -- Pasamos a cada individuo de la generacion 0 por la funcion de coste
        if self.verbose:
            self.IT.header_print(f"Generacion 0")
            self.IT.sub_intro_print("Ejecutando funcion objetivo en los individuos.....")

        for individual in self.population.populuation_dict[str(0)]:
            individual.add_or_update_variable("individual_fitness", self.objective_function(individual))

        if self.verbose:
            self.IT.info_print("Funcion objetivo ejecutada correctamente")
            self.print_generation_info(self.population.populuation_dict[str(0)], 0)

        # </editor-fold>

        # <editor-fold desc="GENERACIONES: REPRODUCCIÓN Y MUTACIONES GENÉTICAS">

        self.IT.header_print(f"Evolución genética de la población ---> ACTIVADA")

        # -- Iteramos por cada generación empezando en la 1
        for gen in range(1, self.num_generations):

            self.IT.header_print(f"Empezando con la generación {gen}")

            # <editor-fold desc="DESARROLLO DEL TORNEO">

            # -- Pintamos a los ganadores de la generación anterior
            if self.verbose:
                self.IT.sub_intro_print(f"Ganadores de la generacion {gen}")

            # -- Ejecutamos el torneo para obtener los padres ganadores en base a los individuos de la generacion anterior
            winners_list: List[Individual] = self.GTM.run_tournament(individuals_list=self.population.populuation_dict[str(gen - 1)])

            if self.verbose:
                for winner in winners_list:

                    if self.problem_type == "bounds_restricted":
                        self.IT.info_print(f"ID: {winner.get_individual_id()} --> "
                                           f"WINNER: {winner.get_individual_values()} --> "
                                           f"FITNESS: {winner.get_individual_fitness()} --> "
                                           f"PARAMETROS VQC: {winner.get_vqc_parameters_values()}")

                    else:
                        self.IT.info_print(f"ID: {winner.get_individual_id()} --> "
                                           f"WINNER: {winner.get_individual_values()} --> "
                                           f"FITNESS: {winner.get_individual_fitness()}")


            # </editor-fold>

            # <editor-fold desc="REPRODUCCIÓN CUÁNTICA DE NUEVOS INDIVIDUOS">

            self.IT.sub_intro_print("Etapa de reproducción cuántica de nuevos individuos")

            # -- Creamos los hijos y los agregamos a la lista de individuos
            children_list: List[Individual] = Reproduction(winners_list=winners_list,
                                                           number_of_children=self.num_individuals,
                                                           problem_restrictions=self.problem_restrictions,
                                                           problem_type=self.problem_type,
                                                           return_to_origin=self.return_to_origin,
                                                           metric_to_optimize=self.metric_to_optimize,
                                                           verbose=self.verbose).run_reproduction(quantum_technology=self.optimization_quantum_technology,
                                                                                                  service=self.optimization_service,
                                                                                                  qm_api_key=self.qm_api_key,
                                                                                                  qm_connection_service=self.qm_connection_service,
                                                                                                  quantum_machine=self.quantum_machine,
                                                                                                  generations_fitness_statistics_df=self.population.generations_fitness_statistics_df,
                                                                                                  max_qubits=self.max_qubits)



            # </editor-fold>

            # <editor-fold desc="EVALUACIÓN DE CAMBIO DE POLÍTICA DE MUTACIÓN Y EJECUCIÓN">

            self.IT.sub_intro_print("Etapa de evaluación de cambio de política de mutación y ejecución de mutaciones")

            # -- Evaluamos si el modelo ha quedado en un minimo local, en caso afirmativo le damos una explosion de variabilidad
            m_proba, m_gen_proba, m_policy, early_stopping_generations_executed = self.VEM.evaluate_early_stopping(self.population.generations_fitness_statistics_df)

            # -- En caso de que una de las variables retornadas por evaluate_early_stopping sea None, todas son None
            if m_proba is not None:
                self.mutate_probability = m_proba
                self.mutate_gen_probability = m_gen_proba
                self.mutation_policy = m_policy
                self.early_stopping_generations_executed = early_stopping_generations_executed

            # -- En caso de darse las condiciones, inicia el CRAZY MODE
            if self.early_stopping_generations_executed and self.mutation_policy == "hard":
                self.IT.header_print("CRAZY MODE ON", "light_red")
                self.variability_explosion_starts_in_generation = gen

            # -- Mutamos los individuos
            children_list = Mutation(children_list, self.mutate_probability, self.mutate_gen_probability, self.mutation_policy, self.problem_restrictions, self.num_generations).run_mutation()

            # -- Agregamos los individuos al diccionario de poblacion en su generacion correspondiente
            self.population.add_generation_population(children_list, gen)

            # </editor-fold>

            # <editor-fold desc="EJECUCIÓN DE FUNCIÓN OBJETIVO">

            # -- Pasamos a cada individuo de la gen=gen por la funcion de coste
            self.IT.sub_intro_print("Ejecutando funcion objetivo en los individuos.....")

            for individual in self.population.populuation_dict[str(gen)]:
                if not individual.get_individual_malformation():
                    individual.add_or_update_variable("individual_fitness", self.objective_function(individual))

            # </editor-fold>

            # <editor-fold desc="ESTADÍSTICA GENERAL DE LA GENERACIÓN">

            self.IT.sub_intro_print("Etapa estadística de la generación")
            self.print_generation_info(self.population.populuation_dict[str(gen)], gen)

            if self.verbose:

                best_fitness = None
                for gen_n, ind_list in self.population.populuation_dict.items():
                    best_gen_ind = sorted(
                        [ind for ind in self.population.populuation_dict[gen_n] if ind.get_individual_fitness() is not None],
                        key=lambda ind: ind.get_individual_fitness(),
                        reverse=self.problem_type == 'maximize'
                    )[0]

                    self.IT.info_print(f"Mejor ind gen: {gen_n}: {best_gen_ind.get_individual_values()} - Fitness: {best_gen_ind.get_individual_fitness()}")

                    if self.problem_type == "maximize":
                        if best_fitness is None:
                            best_fitness = best_gen_ind.get_individual_fitness()
                        else:
                            if best_fitness < best_gen_ind.get_individual_fitness():
                                best_fitness = best_gen_ind.get_individual_fitness()
                    else:
                        if best_fitness is None:
                            best_fitness = best_gen_ind.get_individual_fitness()
                        else:
                            if best_fitness > best_gen_ind.get_individual_fitness():
                                best_fitness = best_gen_ind.get_individual_fitness()

                self.IT.info_print(f"Mejor ind TOTAL: Fitness: {best_fitness}", "light_magenta")

            # -- Evaluamos si despues de la explosion de variabilidad, han transcurrido las generaciones de margen. En caso afirmativo, salimos del bucle
            if self.VEM.stop_genetic_iterations(self.population.generations_fitness_statistics_df):
                break

            # </editor-fold>

        self.IT.header_print("POBLACIÓN FINAL")

        match self.problem_restrictions:

            case "bound_restricted":

                # -- Iteramos por generacion
                for generation, pops in self.population.populuation_dict.items():

                    if self.verbose:
                        self.IT.sub_intro_print(f"GENERACION: {generation}")

                    if self.verbose:

                        for individual in pops:
                            self.IT.info_print(f"ID: {individual.get_individual_id()} --> "
                                               f"VALORES: {individual.get_individual_values()} --> "
                                               f"ANGULOS: {individual.get_vqc_parameters_values()} --> "
                                               f"FITNESS: {individual.get_individual_fitness()}")

                    # -- Mostramos el mejor individuo de la generación
                    if pops:

                        match self.problem_type:
                            case "maximize":
                                self.best_individual_values = max(pops, key=lambda ind: ind.get_individual_fitness())
                            case "minimize":
                                self.best_individual_values = min(pops, key=lambda ind: ind.get_individual_fitness())
                            case _:
                                self.best_individual_values = None  # por si se define otro tipo no soportado

                        if self.best_individual_values:
                            self.IT.info_print(
                                f"MEJOR INDIVIDUO GENERACION: {generation} --> "
                                f"ID {self.best_individual_values.get_individual_id()} --> "
                                f"VALORES: {self.best_individual_values.get_individual_values()} --> "
                                f"ANGULOS: {self.best_individual_values.get_vqc_parameters_values()} --> "
                                f"FITNESS: {self.best_individual_values.get_individual_fitness()}", "light_magenta")

            case "totally_restricted":

                # -- Iteramos por generacion
                for generation, pops in self.population.populuation_dict.items():

                    # -- Mostramos el mejor individuo de la generación
                    if pops:

                        match self.problem_type:
                            case "maximize":
                                self.best_individual_values = max(pops, key=lambda ind: ind.get_individual_fitness())
                            case "minimize":
                                self.best_individual_values = min(pops, key=lambda ind: ind.get_individual_fitness())
                            case _:
                                self.best_individual_values = None  # por si se define otro tipo no soportado

                        if self.best_individual_values:
                            self.IT.info_print(
                                f"MEJOR INDIVIDUO GENERACION: {generation} --> "
                                f"ID {self.best_individual_values.get_individual_id()} --> "
                                f"VALORES: {self.best_individual_values.get_individual_values()} --> "
                                f"FITNESS: {self.best_individual_values.get_individual_fitness()}", "light_magenta")


        # </editor-fold>

    # <editor-fold desc="GETTER DE LOS VALORES DEL MEJOR INDIVIDUO">
    def get_best_individual_values(self):
        """
        Getter para retornar el mejor individuo
        :return:
        """
        return self.best_individual_values.get_individual_values()

    # </editor-fold>

    # <editor-fold desc="MÉTODOS AUXILIARES DE VALIDACIÓN, TORNEOS, CÁLCULOS ESTADÍSTICOS Y PLOTEO">

    def validate_input_parameters(self) -> bool:
        """
        Metodo para validar los inputs que se han cargado en el constructor
        :return: True si todas las validaciones son correctas Excepction else
        """

        # <editor-fold desc="Validamos Dicts" >
        if self.bounds_dict is None:
            raise ValueError("bounds_dict no existe -> None")

        # -- INTERVAL
        interval_bounds_dict: dict = {k: v for k, v in self.bounds_dict.items() if v["bound_type"] == "interval"}
        if not all(isinstance(valor, (int, float)) for param in interval_bounds_dict for key in ["limits", "malformation_limits"] if key in param for valor in param[key]):
            raise ValueError("bounds_dict: No todos los valores en los bounds_dict interval son int o float.")

        # -- PREDEFINED
        predefined_bounds_dict: dict = {k: v for k, v in self.bounds_dict.items() if v["bound_type"] == "predefined"}
        if not all(isinstance(valor, (int, float)) for param in predefined_bounds_dict for key in ["limits", "malformation_limits"] if key in param for valor in param[key]):
            raise ValueError("bounds_dict: No todos los valores en los bounds_dict interval son int o float.")

        if self.element_matrix is not None:
            if not all(isinstance(k, int) and isinstance(v, tuple) and len(v) >= 1 and all(isinstance(coord, (int, float)) for coord in v) for k, v in self.element_matrix.items()):
                raise ValueError("element_matrix: Las claves deben ser int y los valores deben ser tuplas con al menos un número (int o float).")

        # </editor-fold>

        # <editor-fold desc="Validamos callables" >
        # -- objective_function
        if not isinstance(self.objective_function, Callable):
            raise ValueError(f"self.objective_function: Debe ser Callable y su tipo es {type(self.objective_function)}")

        # </editor-fold>

        # <editor-fold desc="Validamos enteros">

        # -- num_generations, num_individuals, podium_size, early_stopping_generations, early_stopping_generations_executed_counter
        if not isinstance(self.num_generations, int):
            raise ValueError(f"self.num_generations: Debe ser un entero y su tipo es {type(self.num_generations)}")
        if not isinstance(self.num_individuals, int):
            raise ValueError(f"self.num_individuals: Debe ser un entero y su tipo es {type(self.num_individuals)}")
        if not isinstance(self.podium_size, int):
            raise ValueError(f"self.podium_size: Debe ser un entero y su tipo es {type(self.podium_size)}")
        if not isinstance(self.early_stopping_generations, int):
            raise ValueError(f"self.early_stopping_generations: Debe ser un entero y su tipo es {type(self.early_stopping_generations)}")
        if not isinstance(self.early_stopping_generations_executed_counter, int):
            raise ValueError(f"self.early_stopping_generations: Debe ser un entero y su tipo es {type(self.early_stopping_generations_executed_counter)}")
        if not isinstance(self.variability_round_decimals, int):
            raise ValueError(f"self.variability_round_decimals: Debe ser un entero y su tipo es {type(self.variability_round_decimals)}")

        # </editor-fold>

        # <editor-fold desc="Validamos flotantes">

        # -- mutate_probability, mutate_gen_probability
        if not isinstance(self.mutate_probability, float):
            raise ValueError(f"self.mutate_probability: Debe ser un float y su tipo es {type(self.mutate_probability)}")
        if not isinstance(self.mutate_gen_probability, float):
            raise ValueError(f"self.mutate_gen_probability: Debe ser un float y su tipo es {type(self.mutate_gen_probability)}")

        # </editor-fold>

        # <editor-fold desc="Validamos Literals/strings">

        # -- problem_type, problem_restrictions, tournament_method
        if not isinstance(self.problem_type, str):
            raise ValueError(f"self.problem_type: Debe ser un str y su tipo es {type(self.problem_type)}")
        if self.problem_type not in ["minimize", "maximize"]:
            raise ValueError(f'self.problem_type debe ser una opción de estas: ["minimize", "maximize"] y se ha pasado {self.problem_type}')

        if self.problem_restrictions != "bound_restricted":
            if not isinstance(self.return_to_origin, str) or self.return_to_origin is None:
                raise ValueError(f"self.problem_type: Debe ser un str y su tipo es {type(self.return_to_origin)}")

        if self.problem_restrictions != "bound_restricted":
            if self.return_to_origin not in ["return_to_origin", "no_return"] and self.problem_type is not None:
                raise ValueError(f'self.problem_type debe ser una opción de estas: ["return_to_origin", "return_to_origin"] y se ha pasado {self.return_to_origin}')
        if not isinstance(self.problem_restrictions, str):
            raise ValueError(f"self.problem_restrictions: Debe ser un str y su tipo es {type(self.problem_restrictions)}")
        if self.problem_restrictions not in ["bound_restricted", "totally_restricted"]:
            raise ValueError(f'self.problem_restrictions debe ser una opción de estas: ["bound_restricted", "totally_restricted"] y se ha pasado {self.problem_restrictions}')
        if not isinstance(self.mutation_policy, str):
            raise ValueError(f"self.mutation_policy: Debe ser un str y su tipo es {type(self.mutation_policy)}")
        if self.mutation_policy not in ["soft", "normal", "hard"]:
            raise ValueError(f'self.mutation_policy debe ser una opción de estas: ["soft", "normal", "hard"] y se ha pasado {self.mutation_policy}')
        if not isinstance(self.tournament_method, str):
            raise ValueError(f"self.tournament_method: Debe ser un str y su tipo es {type(self.tournament_method)}")

        # </editor-fold>>

        # <editor-fold desc="Validamos bools">

        # -- verbose, early_stopping_generations_executed
        if not isinstance(self.verbose, bool):
            raise ValueError(f"self.verbose: Debe ser un bool y su tipo es {type(self.verbose)}")
        if not isinstance(self.early_stopping_generations_executed, bool):
            raise ValueError(f"self.early_stopping_generations_executed: Debe ser un bool y su tipo es {type(self.early_stopping_generations_executed)}")

        # </editor-fold>

        return True

    def get_tournament_method(self, verbose):
        """
        Metodo que crea y retorna el tournament seleccionado
        :return:
        """
        match self.tournament_method:

            case "ea_simple":
                self.IT.info_print("Se ha configurado existosamente el torneo ea_simple")
                return EaSimple(self.podium_size, self.problem_type, verbose)

    def print_generation_info(self, individual_generation_list: List[Individual], generation: int) -> None:
        """
        Metodo para pintar información de la generación por consola
        :param individual_generation_list: [List[Individual]] Lista de individuos
        :param generation: [int] Número de la generación
        :return: None
        """
        self.IT.sub_intro_print("Información de los individuos y los fitness")
        for i, ind in enumerate([z for z in individual_generation_list if z.generation == generation]):
            self.IT.info_print(f"Individuo {ind.get_individual_id()}: {ind.get_individual_values()} - Generación: {ind.generation} - [Fitness]: {ind.get_individual_fitness()}")

        self.IT.sub_intro_print(f"Información de la evolución de las distribuciones en cada generación")
        self.IT.print_tabulate_df(self.population.get_generation_fitness_statistics(generation), row_print=self.num_generations+1)

        return None

    def plot_generation_stats(self) -> None:
        self.population.plot_generation_stats(self.variability_explosion_starts_in_generation)

    def plot_evolution_animated(self) -> None:
        self.population.plot_evolution_animated()

    def plot_evolution(self) -> None:
        self.population.plot_evolution()

    # </editor-fold>

__all__ = ['GenethicOptimizer']